#include <FlexiTimer2.h>

#include "EasyTransfer.h"
#include "EasyFunctions_V2.h"
#include "stdlib.h"
#include <Servo.h> 
#include "math.h"

#include "digitalWriteFast.h"


#define S_TOP 2
#define S_LEFT 3 
#define S_RIGHT 4 
#define S_BOTTOM 5 


float test;

/* Communication protocol :
   The master transmits data with the enum type order to indicate how to handle it, what action to take
   The slave only needs to transmit current position from the odometry at certain times, in this case order will be set to SLAVE_TRANSMISION and the master will recover the data
   Stocked by default as : data1 : x , data2 : y , data3 : angle 
   If the slave has completed a task and is ready for a new one, it can transmit AWAITING_ORDER
*/
//create object
EasyTransfer ET; 


#define NB_TICKS_FULL_ROTATION 		2048
#define RADIUS_ENCODER_WHEEL		0.03
#define SPACING_ENCODERS		0.23

DATA_STRUCTURE mydata;


// encoder right
#define encInterrPinARigth 			0
#define encInterrPinBRigth 			1
#define encPinARight 			2
#define encPinBRight 			3
// encoder left
#define encInterrPinALeft 			2
#define encInterrPinBLeft 			3
#define encPinALeft 			21
#define encPinBLeft 			20


#define DT_MILLIS			300
#define DT_SEC				0.3
int counter;



 volatile ENCODER_INFO encInfo;



/*       END ENCODER.H copy*/


/**
    ENCODER COPY :
    */
    
void enc_updt_estimation(){

	//estimation of the distance covered by each wheel since the last reset (integral)
	encInfo.disCovR -= encInfo.ticksR;
	encInfo.disCovL -= encInfo.ticksL;
	mydata.data1=encInfo.disCovL;
        mydata.data2=encInfo.disCovR;
        
        //reset the ticks
        encInfo.ticksL=0;
        encInfo.ticksR=0;
}

void enc_reset(){
	encInfo.ticksR = 0;
	encInfo.ticksL = 0;
	
	encInfo.disCovR = 0;
	encInfo.disCovL = 0;
}


// interrution routines attached the the various encoders:
// interruption routine attached to the A pin of the right encoder
void interruptFunctionEncPinARight(){
	if (digitalReadFast2(encPinARight) == digitalReadFast2(encPinBRight)) {
		encInfo.ticksR--;
	}
	else {
		encInfo.ticksR++;
	}
}
 
// interruption routine attached to the  pin of the right encoder
void interruptFunctionEncPinBRight(){
	if (digitalReadFast2(encPinARight) == digitalReadFast2(encPinBRight)) {
		encInfo.ticksR++;
	}
	else {
		encInfo.ticksR--;
	}
}

// interruption routine attached to the A pin of the left encoder
void interruptFunctionEncPinALeft(){
	if (digitalReadFast2(encPinALeft) == digitalReadFast2(encPinBLeft)) {
		encInfo.ticksL--;
	}
	else {
		encInfo.ticksL++;
	}
}
 
// interruption routine attached to the B pin of the left encoder
void interruptFunctionEncPinBLeft(){
	if (digitalReadFast2(encPinALeft) == digitalReadFast2(encPinBLeft)) {
		encInfo.ticksL++;
	}
	else {
		encInfo.ticksL--;
	}
}
    

/* END OF ENCODER COPY  */



/* AUTO.H copy */

// contantes
#define R_WEEL			30 
#define R_ENC			30
#define SPACING_	
#define SPACING_ENC		230

#define STOP_DISTANCE		200

#define DT			50

//limit maximum speed to avoid going to fast in either direction
#define MAXIMUM_MOTOR_VELOCITY	200
#define MINIMUM_MOTOR_VELOCITY	50


//////////////////////////////////////////////////////////////////////////////////////////
// AUTO
//////////////////////////////////////////////////////////////////////////////////////////


volatile AUTO_INFO autoInfo;


//////////////////////////////////////////////////////////////////////////////////////////
// PID
//////////////////////////////////////////////////////////////////////////////////////////

#define KP_PID_DIST_L	3
#define KI_PID_DIST_L	0
#define KD_PID_DIST_L 	0

#define KP_PID_DIST_R   1
#define KI_PID_DIST_R	0
#define KD_PID_DIST_R	0

#define KP_PID_VEL_L	3
#define KI_PID_VEL_L	0
#define KD_PID_VEL_L	0

#define KP_PID_VEL_R 	1
#define KI_PID_VEL_R	0
#define KD_PID_VEL_R	0
/*

#define KP_PID_VEL_L	1.66666666667
#define KI_PID_VEL_L	5
#define KD_PID_VEL_L	0.05

#define KP_PID_VEL_R 	1.66666666667
#define KI_PID_VEL_R	5
#define KD_PID_VEL_R	0.05
*/


PID pid_dist_L = {KP_PID_DIST_L, KI_PID_DIST_L, KD_PID_DIST_L, DT_SEC};
PID pid_dist_R = {KP_PID_DIST_R, KI_PID_DIST_R, KD_PID_DIST_R, DT_SEC};
PID pid_vel_L = {KP_PID_VEL_L, KI_PID_VEL_L, KD_PID_VEL_L, DT_SEC};
PID pid_vel_R = {KP_PID_VEL_R, KP_PID_VEL_R, KD_PID_VEL_R, DT_SEC};
/* END OF AUTO.H copy */



/* AUTO.cpp copy */

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// getters and setters
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


// updates the aimed position, save also the starting position of the motion
void auto_updt_aimed_position(int newCommandx, int newCommandy, float newCommandPhi){
	// first : saving the starting position of the new motion
	auto_updt_initial_position(autoInfo.currentX, autoInfo.currentY, autoInfo.currentPhi);
	
	// update the command
	autoInfo.commandX=newCommandx;
	autoInfo.commandY=newCommandy;
	autoInfo.commandPhi=newCommandPhi;
	autoInfo.commandDistance = sqrt((autoInfo.commandX-autoInfo.x0)*(autoInfo.commandX-autoInfo.x0)
					+(autoInfo.commandY-autoInfo.y0)*(autoInfo.commandY-autoInfo.y0));
}

// updates the initial position of the motion : velocity profile generation
void auto_updt_initial_position(int newX0, int newY0, float newPhi0){
	autoInfo.x0=newX0;
	autoInfo.y0=newY0;
	autoInfo.phi0=newPhi0;
}

long dist_to_ticks(float dist){
   return (dist*4096)/(2*PI*RADIUS_ENCODER_WHEEL);
}


////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Command calculation
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// updates the command applied to the motor, launch after the update of the current position
void auto_updt_command(volatile ENCODER_INFO *encInfo){
  
        
	float comWheelFromAng = autoInfo.commandPhi*SPACING_ENCODERS/2;
	int DCommandLeft = dist_to_ticks(autoInfo.commandDistance-comWheelFromAng);
	int DCommandRight = dist_to_ticks(autoInfo.commandDistance+comWheelFromAng);
	
	//left motor
	//distance error
	int DErrorLeft = DCommandLeft-encInfo->disCovL;
        pid_updt(&pid_dist_L, DErrorLeft);
        mydata.data3=2*PI*DCommandLeft/(4096*RADIUS_ENCODER_WHEEL);

	// velocity error
	int VErrorLeft = pid_dist_L.out-encInfo->ticksL;
	pid_updt(&pid_vel_L, VErrorLeft);

	autoInfo.leftMotorCommand=pid_vel_L.out;
	

	//right motor
	//distance error
	int DErrorRight = DCommandRight-encInfo->disCovR;
	pid_updt(&pid_dist_R, DErrorRight);
        mydata.data4=2*PI*DCommandRight/(4096*RADIUS_ENCODER_WHEEL);
	// velocity error
	int VErrorRight = pid_dist_R.out-encInfo->ticksR;
	pid_updt(&pid_vel_R, VErrorRight);

	autoInfo.rightMotorCommand=pid_vel_R.out;
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// Command command application
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// applies the command to the system (motors)
void auto_appl_command(){

        //offset L298 128 => stay still
        autoInfo.rightMotorCommand += 128;
        autoInfo.leftMotorCommand += 128;
	// normalisation
       /* if( (autoInfo.commandDistance - autoInfo.commandPhi*SPACING_ENCODERS/2) - (encInfo.disCovR)>0.001){// && autoInfo.rightMotorCommand<152){
          autoInfo.rightMotorCommand=148;
        }
        if( (autoInfo.commandDistance - autoInfo.commandPhi*SPACING_ENCODERS/2) - (encInfo.disCovL) >0.001 ){//&& autoInfo.leftMotorCommand<148){
          autoInfo.leftMotorCommand=148;
        }*/
	autoInfo.rightMotorCommand = (autoInfo.rightMotorCommand <= MAXIMUM_MOTOR_VELOCITY)? autoInfo.rightMotorCommand : MAXIMUM_MOTOR_VELOCITY;
	autoInfo.rightMotorCommand = (autoInfo.rightMotorCommand >= MINIMUM_MOTOR_VELOCITY)? autoInfo.rightMotorCommand : MINIMUM_MOTOR_VELOCITY;
	
	autoInfo.leftMotorCommand = (autoInfo.leftMotorCommand <= MAXIMUM_MOTOR_VELOCITY)? autoInfo.leftMotorCommand : MAXIMUM_MOTOR_VELOCITY;
	autoInfo.leftMotorCommand = (autoInfo.leftMotorCommand >= MINIMUM_MOTOR_VELOCITY)? autoInfo.leftMotorCommand : MINIMUM_MOTOR_VELOCITY;

	//mydata.data1=autoInfo.leftMotorCommand;
        //mydata.data2=autoInfo.rightMotorCommand;

        
	// application
	PWM_Left_Motor((int)autoInfo.leftMotorCommand);
	PWM_Right_Motor((int)autoInfo.rightMotorCommand);
	return;
}

// Correspond au moteur 2 sur la carte (à 31khz)
void PWM_Left_Motor(int direct){
    // direct : valeur entre 0 et 255
    // envoi sur le pin 9 la valeur de direct, et sur le pin 10 son opposé
  //COM2 A1 and A0 at 1 for inverse PWM mode, just B1 at 1 for normal PWM
  // |= to not modify existing bit conditions on other registers
  TCCR2A |= _BV(COM2A1) | _BV(COM2A0) | _BV(COM2B1) | _BV(WGM20); // 8 bit phase correct PWM mode
  TCCR2B |= _BV(CS20);// prescale = 1 CS2:0
  TCCR2B = TCCR2B & 0b11111001;
  OCR2A=direct;
  OCR2B=direct;
  }
  
  // correspond au moteur 1 sur la carte (à 31kHz)
void PWM_Right_Motor(int direct){
    // direct : value between 0 et 255t
    // envoi sur le pin 12 la valeur de direct, et sur le pin 11 son opposé
    //COM2 A1 and A0 at 1 for inverse PWM mode, just B1 at 1 for normal PWM
    TCCR1A |= _BV(COM1A1) | _BV(COM1A0) | _BV(COM1B1) | _BV(WGM20); // 8 bit phase correct PWM mode
    TCCR2B |= _BV(CS20);// prescale = 1 CS2:0
    TCCR1B = TCCR1B & 0b11111001;
    OCR1A=direct;
    OCR1B=direct;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// PID
/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
void pid_updt(PID *pid, float error){
	int derive = (error-pid->previousError)/pid->T;
	int integral = pid->previousIntegral + (error-pid->previousError)/2.*pid->T;
	pid->out = pid->Kp*error + pid->Ki*integral + pid->Kd*derive;
	
	pid->previousError = error;
	pid->previousIntegral = integral;
}




void wait_for(int time){
        long tBegin=millis();
        while(millis()<tBegin+time){
        }
}

//Create data structure
float old;

void setup(){
  Serial1.begin(9600);
  //start the library, pass in the data details and the name of the serial port. Can be Serial, Serial1, Serial2, etc.
  ET.begin(details(mydata), &Serial1);

  
  
  /** ENCODERS */
  // Encoders
  
	// Right encoder
  	pinMode(encPinARight, INPUT);      // input pin A
	pinMode(encPinBRight, INPUT);      // input pin B
  	digitalWrite(encPinARight, HIGH);  // pullup
	digitalWrite(encPinBRight, HIGH);  // pullup
  	attachInterrupt(encInterrPinARigth, interruptFunctionEncPinARight, RISING); // 4096 ticks/s (4*1024)
  	attachInterrupt(encInterrPinBRigth, interruptFunctionEncPinBRight, RISING);
 
  	// Left encoder
  	pinMode(encPinALeft, INPUT);
  	pinMode(encPinBLeft, INPUT);
  	digitalWrite(encPinALeft, HIGH);
  	digitalWrite(encPinBLeft, HIGH);
  	attachInterrupt(encInterrPinALeft, interruptFunctionEncPinALeft, RISING);
  	attachInterrupt(encInterrPinBLeft, interruptFunctionEncPinBLeft, RISING);
    
        counter=0;  
  // initialise servos
    //pinMode(13, OUTPUT);

  //servoNet.attach(SERVO_NET);
  
  
  // Put initial values to the components of the data structure
  mydata.data1=1;
  mydata.current1=NO_DATA1;
  mydata.current2=NO_DATA2;
  
  pinMode(9, OUTPUT);
  pinMode(10, OUTPUT);

  pinMode(11, OUTPUT);

  pinMode(12, OUTPUT);
  mydata.data2=2;
  old=0;
  	// isrt
	FlexiTimer2::set(DT_MILLIS, 1/1000., isrt);
  	FlexiTimer2::start();
}

void isrt(){
	//reading the sensors 
	enc_updt_estimation();
	//calculation of the command
	auto_updt_command(&encInfo);
	//application of the command on the system (motors)
        auto_appl_command();


}
void loop(){


  autoInfo.commandPhi=0;
  autoInfo.commandDistance=0.3;
  while(1){
     
           ET.sendData();
           wait_for(500); 
          
        

  }
    
  
     
  }  


